/*!
 * @file element_2d.h
 * Definition of the 2D elements rendered on top through the GraphicsEngine
 * @todo reimplement it, so it looks just like PNode.
*/

#ifndef _ELEMENT_2D_H
#define _ELEMENT_2D_H

#include "base_object.h"
#include "vector.h"

// FORWARD DECLARATION
class PNode;
class TiXmlElement;
template<class T> class tList;

//!< An enumerator defining the Depth of a 2D-element.
typedef enum
{
  E2D_BELOW_ALL                 =     1,        //!< Will be rendered below the 3D-scene. @todo make this work.
  E2D_BOTTOM                    =     2,        //!< Will be rendered on the bottom Layer
  E2D_MEDIUM                    =     4,        //!< Will be rendered on the medium Layer.
  E2D_TOP                       =     8,        //!< Will be rendered on top of everything else

  E2D_LAYER_COUNT               =     4         //!< The count of Layers.
} E2D_LAYER;
#define E2D_DEFAULT_LAYER       E2D_TOP
#define E2D_ALL_LAYERS          E2D_TOP | E2D_MEDIUM | E2D_BOTTOM | E2D_BELOW_ALL

typedef enum
{
  E2D_ALIGN_NONE                =     0,
  E2D_ALIGN_LEFT                =     1,
  E2D_ALIGN_RIGHT               =     2,
  E2D_ALIGN_CENTER              =     4,
  E2D_ALIGN_SCREEN_CENTER       =     8
} E2D_ALIGNMENT;

typedef enum
{
  E2D_LOCAL_ROTATE          =   1,    //!< Rotates all the children around their centers.
  E2D_ROTATE_MOVEMENT       =   2,    //!< Moves all the children around the center of their parent, without the rotation around their own centers.

  E2D_MOVEMENT              =   4,    //!< Moves all children along with the parent.
// special linkage modes
  E2D_ALL                   =   3,    //!< Moves all children around the center of their parent, and also rotates their centers
  E2D_ROTATE_AND_MOVE       =   5     //!< Rotates all children around their axis, and moves them as the Parent Moves, but does not rotate around the center of their parent.
} E2D_PARENT_MODE;
#define     E2D_DEFAULT_PARENTING_MODE  E2D_ALL

//! A Struct defining the Position of an Element in 2D-space
struct Position2D
{
  float       x;                 //!< The x-coordinate.
  float       y;                 //!< The y-coordinate.
  float       depth;             //!< The distance from the viewing plane.
};

//! A class for ...
class Element2D : virtual public BaseObject {

  public:
    Element2D();
    virtual ~Element2D();

    void init();
    void loadParams(const TiXmlElement* root);

    /** @param xCoord the xCoordinate @param yCoord the y-Coordinate. These coordinates are Relative */
    inline void setPosition2D(int xCoord, int yCoord) { this->relPos2D[0] = xCoord; this->relPos2D[1] = yCoord; };
    /** this returns the Absolute Position on the screen set by positioning in the tick-phase */
    inline const Position2D& getPosition2D() { return this->absPos2D; };

    /** @param alignment the new Alignment of the 2D-Element */
    inline void setAlignment(E2D_ALIGNMENT alignment) { this->alignment = alignment; };
    void setAlignment(const char* alignment);

    void setLayer(E2D_LAYER layer);
    void setLayer(const char* layer);
    /** @returns the Layer this Element is drawn to */
    inline E2D_LAYER getLayer() { return this->layer; };

    /** @param visible true if the Element should be visible false otherwise (will not be rendered) */
    inline void setVisibility(bool visible) { this->visible = visible; };
    /** @param active true if the Element should be active, false otherwise (will not be ticked) */
    inline void setActiveness(bool active) { this->active = active; };


    /** @param bindNode the Node this 2D-element should follow. if NULL the Element will not follow anything */
    inline void setBindNode(const PNode* bindNode) { this->bindNode = bindNode; };
    void setBindNode(const char* bindNode);

    /** @returns the visibility state */
    inline bool isVisible() { return this->visible; };
    /** @returns the active-State */
    inline bool isActive() { return this->active; };


    // LIKE PNODE
  public:
    void setRelCoor2D (const Vector& relCoord);
    void setRelCoor2D (float x, float y, float dontCare);
    void setRelCoorSoft2D(const Vector& relCoordSoft, float bias = 1.0);
    void setRelCoorSoft2D(float x, float y, float dontCare, float bias = 1.0);
    /** @returns the relative position */
    inline const Vector& getRelCoor2D () const { return this->prevRelCoordinate; };
    void setAbsCoor2D (const Vector& absCoord);
    void setAbsCoor2D (float x, float y, float depth);
    /** @returns the absolute position */
    inline const Vector& getAbsCoor2D () const { return this->absCoordinate; };
    void shiftCoor (const Vector& shift);

    void setRelDir2D (float relDir);
    void setRelDirSoft2D(float relDirSoft, float bias = 1.0);
    /** @returns the relative Direction */
    inline float getRelDir2D () const { return this->prevRelDirection; };
    void setAbsDir2D (float absDir);
    /** @returns the absolute Direction */
    inline float getAbsDir2D () const { return this->absDirection; };
    void shiftDir (float shiftDir);

    /** @returns the Speed of the Node */
    inline float getSpeed() const { return 0; };
    /** @returns the Velocity of the Node */
    inline const Vector& getVelocity() const { return this->velocity; };


    void addChild2D (Element2D* child, int parentingMode = E2D_DEFAULT_PARENTING_MODE);
    void addChild2D (const char* childName);
    void removeChild2D (Element2D* child);
    void remove2D();

    void setParent2D (Element2D* parent);
    void setParent2D (const char* parentName);
    /** @returns the parent of this Element2D */
    Element2D* getParent () const { return this->parent; };

    void softReparent(PNode* parentNode, float bias = 1.0);
    void softReparent(const char* parentName, float bias = 1.0);

    /** @param parentMode sets the parentingMode of this Node */
    void setParentMode2D (E2D_PARENT_MODE parentMode) { this->parentMode = parentMode; };
    void setParentMode2D (const char* parentingMode);
    /** @returns the Parenting mode of this node */
    int getParentMode2D() const { return this->parentMode; };

    void update2D (float dt);

    void debug (unsigned int depth = 1, unsigned int level = 0) const;
    void debugDraw2D(unsigned int depth = 1, float size = 1.0, Vector color = Vector(1,1,1)) const;

    // helper functions //
    static const char* parentingModeToChar(int parentingMode);
    static E2D_PARENT_MODE charToParentingMode(const char* parentingMode);

  private:
    void init(Element2D* parent);
    /** tells the child that the parent's Coordinate has changed */
    inline void parentCoorChanged () { this->bRelCoorChanged = true; }
    /** tells the child that the parent's Direction has changed */
    inline void parentDirChanged () { this->bRelDirChanged = true; }
    /** @returns the last calculated coordinate */
    inline Vector getLastAbsCoor() { return this->lastAbsCoordinate; }

  public:
    virtual void tick(float dt);
    virtual void draw() const = NULL;

  protected:
    void positioning();

  protected:
    const PNode*            bindNode;         //!< a node over which to display this 2D-element
    int                     relPos2D[2];      //!< X-coord, Y-Coord (relative to the Coordinates of the alignment if given.)
    Position2D              absPos2D;         //!< The absolute position of the 2D-Element.

    E2D_ALIGNMENT           alignment;        //!< How the Element is aligned around its Position

  private:
    bool                    visible;          //!< If the given Element2D is visible.
    bool                    active;           //!< If the given Element2D is active.
    E2D_LAYER               layer;            //!< What layer this Element2D is on.





  private:
    bool              bRelCoorChanged;    //!< If Relative Coordinate has changed since last time we checked
    bool              bRelDirChanged;     //!< If Relative Direction has changed since last time we checked

    Vector            relCoordinate;      //!< coordinates relative to the parent
    Vector            absCoordinate;      //!< absolute coordinates in the world ( from (0,0,0) )
    float             relDirection;       //!< direction relative to the parent
    float             absDirection;       //!< absolute direvtion in the world ( from (0,0,1) )

    Vector            prevRelCoordinate;  //!< The last Relative Coordinate from the last update-Cycle.
    Vector            lastAbsCoordinate;  //!< this is used for speedcalculation, it stores the last coordinate
    float             prevRelDirection;   //!< The last Relative Direciton from the last update-Cycle.
//  Quaternion        lastAbsDirection;

    Vector            velocity;           //!< Saves the velocity.

    Vector*           toCoordinate;       //!< a position to which to iterate. (This is used in conjunction with softReparent.and set*CoorSoft)
    float*            toDirection;        //!< a direction to which to iterate. (This is used in conjunction with softReparent and set*DirSoft)
    float             bias;               //!< how fast to iterate to the given position (default is 1)

    Element2D*        parent;             //!< a pointer to the parent node
    tList<Element2D>* children;           //!< list of the children of this Element2D

    unsigned int      parentMode;         //!< the mode of the binding
};


#endif /* _ELEMENT_2D_H */
